package bank3tier.db; //Pool

import java.util.List;
import bank3tier.db.DbObject.OwnerPoolChangedExc;
import multex.Exc;


/**Operative interface to the persistence layer similar to Siedersleben, page 204, 
 * but with some more operations. 
 * This is the interface, which should be used mostly by the business logic layer.
 * @author Christoph Knabe 2005-03-28
 */
public interface Pool {


    /*package*/ static final Integer createdOid = null;
    /*package*/ static final Integer deletedOid = new Integer(0);

    /**Makes the transient object persistent by inserting it into this Pool.
     * It gets a new OID greater than {@link #deletedOid}.
     * Its Pool is set to this Pool.
     * @throws OwnerPoolChangedExc 
    */
    void insert(DbObject io_object) throws OwnerPoolChangedExc;
    
    /**Reattaches the detached object to this Pool.
     * It will overwrite the corresponding object with the same OID, when the actual transaction is committed.
     * Its Pool must be this Pool.
     * This operation must be called early in a transaction.
     * @throws OwnerPoolChangedExc 
    */
    void reattach(DbObject io_object);
    
    /**Saves the actual state of the object into the pool.
     * If the object was transient before, it gets a new OID greater than {@link #deletedOid}.
     * If it was detached, it is reassociated with the current session and will be stored into the pool on the next commit.
     * Its Pool is set to this Pool.
     * @return true newly inserted into the pool.
     * @throws ConcurrentUpdateExc  Another user modified this object in the Pool before you.
     *                              Thus you should repeat your transaction.
     */
    boolean save(DbObject io_object) throws ConcurrentUpdateExc, OwnerPoolChangedExc;
    
    /**Deletes the object from the pool.
     * Additonally mark the object as not valid by giving it the OID {@link #deletedOid}.
     * @return true it was in the pool.
     * TODO May be ConcurrentUpdateExc, too, if it was updated by another user
     * after loading.
     */
    boolean delete(DbObject io_object);
    
    /**Returns the element of the Pool with the given Class and OID.*/
    <T extends DbObject> T find(Class<T> i_persistentClass, Integer i_oid) throws NoPersistentClassExc, OidNotFoundExc;

    /**Returns exactly one element from the Pool with the given Class.
     * @param i_persistentClass Class of the desired object
     * @param i_keyFieldName The name of the field of the class, which will be used as key
     * @param i_keyFieldValue The unique key value, which serves for searching with the key i_keyFieldName 
     * @return the found object
     * @throws NoPersistentClassExc i_persistentClass is no persistent class
     * @throws Exception No one or more than one objects with this key value were found or another failure
     */    
    <T extends DbObject> T findOneBykey(Class<T> i_persistentClass, String i_keyFieldName, Object i_keyFieldValue) 
    throws NoPersistentClassExc, KeyNotFoundExc, KeyNotUniqueExc;
    
    /**Returns a List with all elements of the given Class in the Pool.*/
    <T extends DbObject> List<T> findAll(Class<T> i_persistentClass) throws NoPersistentClassExc;

    /**Returns all DbObject-s, which have in the 
     * non-null, non-transient attributes an equal value as i_object.
     * The returned objects will be dynamically an instance of the same type
     * as the actual example object.
     * The resultant List may be empty.
     */ 
    //public <T extends DbObject> List<T> findManyByExample(T i_object);
    //It is not possible to implement this function by the Java Persistence API!
    
    /**Returns all DbObject-s of the given class, which have in the given field the given value.
     * @param i_persistentClass the class, of which objects are searched
     * @param i_keyFieldName the name of the field, which is used as search key
     * @param i_keyFieldValue the value, which is searched for the key field
     * @result all matching objects, or an empty List.
     * @throws NoPersistentClassExc i_persistentClass is no persistent class
     */ 
    public <T extends DbObject> List<T> findManyByKey(Class<T> i_persistentClass, String i_keyFieldName, Object i_keyFieldValue)
    throws NoPersistentClassExc;

    /**Returns true, if there is an object of i_linkClass, which is a link between i_left and i_right.
     * @throws NoPersistentClassExc i_linkClass is not a persistent class.*/
    public boolean areLinked(DbObject i_left, Class<? extends DbObject> i_linkClass, DbObject i_right) throws NoPersistentClassExc;

    /**Returns true, if there are objects of the class i_resultClass, which have a reference to i_object. 
   * @throws NoPersistentClassExc i_resultClass or the Class of i_object is not a subclass of bank3tier.db.DbObject
  */
  <T extends DbObject> boolean hasLinks(Class<T> i_resultClass, DbObject i_object) throws LinkFailure, NoPersistentClassExc;
  
    /**Returns all objects of the class i_resultClass, which have a reference to i_object. 
   * @throws NoPersistentClassExc i_resultClass or the Class of i_object is not a subclass of bank3tier.db.DbObject
  */
  <T extends DbObject> List<T> getLinks(Class<T> i_resultClass, DbObject i_object) throws LinkFailure, NoPersistentClassExc;
    
    /**Returns all objects of class i_resultClass from the Pool, 
     * which fulfill the query i_queryClass with the given arguments.
     * @throws NoPersistentClassExc i_resultClass is not a subclass of bank3tier.db.DbObject
     * @throws NoQueryClassExc i_queryClass is not a subclass of bank3tier.db.JoraQuery
     */
    <T extends DbObject> List<T> findManyByQuery(Class<T> i_resultClass, Class i_queryClass, Object[] i_args) 
    throws NoPersistentClassExc, NoQueryClassExc, ArgumentCountExc, ArgumentTypeExc;

    /**Object with OID {0} not found in table "{1}"*/
    public static class OidNotFoundExc extends multex.Exc {
        //TODO Prüfen, ob noch nötig! 07-08-24
        public OidNotFoundExc(final Integer i_oid, final String i_table){
            super("Object with OID {0} not found in table \"{1}\"",
                i_oid, i_table
            );
        }
    }
    
    /**{0}-object with OID {1} has been modified by another user. You should repeat the transaction.*/
    public static class ConcurrentUpdateExc extends multex.Exc {
        public ConcurrentUpdateExc(final String i_tableName, final Integer i_oid){
            super("{0}-object with OID {1} has been modified by another user. "
                + " You should repeat the transaction", i_tableName, i_oid
            );
        }
    }

    /**Class {0} is not a persistent class. Operation impossible.*/
    public static class NoPersistentClassExc extends multex.Exc {
        public NoPersistentClassExc(final Class i_class){
            super("Class {0} is not a persistent class. Operation impossible."
                , i_class.getName()
            );
        }
    }

    /**Example class {0} is no subclass of expected class {1}. Object: {2}*/
    public class UnexpectedExampleClassExc extends Exc {

        public UnexpectedExampleClassExc(final String expectedClassName, final String exampleClassName, final Object exampleObject) {
            super("Example class {0} is no subclass of expected class {1}. Object: {2}", exampleClassName, expectedClassName, exampleObject);
        }

    }
    
    /**Class {0}: No object found with key {1}="{2}"*/
    public static class KeyNotFoundExc extends multex.Exc {
        public KeyNotFoundExc(final Class className, final String i_keyFieldName, final Object i_keyFieldValue){
            super("Class {0}: No object found with key {1}=\"{2}\"", className, i_keyFieldName, i_keyFieldValue);
        }
    }

    /**Class {0}: {3} objects instead of one found with key {1}="{2}"*/
    public static class KeyNotUniqueExc extends multex.Exc {
        public KeyNotUniqueExc(final Class className, final String i_keyFieldName, final Object i_keyFieldValue, final int count){
            super("Class {0}: {3} objects instead of one found with key {1}=\"{2}\"", className, i_keyFieldName, i_keyFieldValue, new Integer(count));
        }
    }
    
    /**Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3}*/
    public static class LinkFailure extends multex.Failure {
        public LinkFailure(final Exception i_cause, final String i_ownClassName, final String i_partnerOidFieldName, final String i_partnerTableName, final Integer i_partnerOid){
            super("Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3} ",
                i_cause, i_ownClassName, i_partnerOidFieldName, i_partnerTableName, i_partnerOid
            );
        }
    }
    
    /**Query Class {0} is not a subclass of query base class {1}.*/
    public static class NoQueryClassExc extends multex.Exc {
        public NoQueryClassExc(final Class i_queryClass, final Class i_baseClass){
            super("Query Class {0} is not a subclass of query base class {1}."
                , i_queryClass.getName(), i_baseClass.getName()
            );
        }
    }
    
    /**Query {0} has been called with {1} arguments, but needs {2}.*/
    public static class ArgumentCountExc extends multex.Exc {
        public ArgumentCountExc(final Class i_queryClass, final int i_providedCount, final int i_neededCount){
            super("Query {0} has been called with {1} arguments, but needs {2}."
                , i_queryClass.getName(), new Integer(i_providedCount), new Integer(i_neededCount)
            );
        }
    }
    
    /**Of query {0} the argument index {1} is of class {2}, but should be of class {3}.*/
    public static class ArgumentTypeExc extends multex.Exc {
        public ArgumentTypeExc(final Class i_queryClass, final int i_argumentIndex, final Class i_argumentClass, final Class i_expectedClass){
            super("Of query {0} the argument index {1} is of class {2}, but should be of class {3}."
                , i_queryClass.getName(), new Integer(i_argumentIndex), i_argumentClass.getName(), i_expectedClass.getName()
            );
        }
    }

}
